/*
 * immobile_kings_in_check.c --- check if an immobile king (or have castled as
 *                               its sole move), is in check during the
 *                               trajectories of other pieces.
 *
 * Copyright (c) 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "common.h"
#include "immobile_kings_in_check.h"

static bool_t
is_king_in_check_for_squares(
    square_t dest,
    square_t* squares,
    unsigned int nb_squares)
{
    unsigned int k;
    for (k = 0; k < nb_squares; k++)
    {
        if (dest == squares[k])
        {
            return TRUE;
        }
    }
    return FALSE;
}

static bool_t
is_king_in_check(
    colour_t colour,
    square_t* bishop_squares,
    unsigned int nb_bishop_squares,
    square_t* rook_squares,
    unsigned int nb_rook_squares,
    square_t* knight_squares,
    unsigned int nb_knight_squares,
    square_t* pawn_promotion_squares,
    unsigned int nb_pawn_promotion_squares)
{
    unsigned int i;
    piece_t* p;

    for (i = 0, p = &ListeSures[i];
	 i < NbSures + NbRestantes;
	 i++, p++)
    {
        int j = 0;
        piece_type_t type = p->typePiece;

        if (p->camp == colour) continue;

        while (j < p->nbDestination)
        {
            square_t dest = p->destinations[j].destination;

            if (j < p->nbDestination - 1)
            {
                if (type == FOU
                    || type == DAME
                    || type == PION
                    || type == ROI)
                {
                    if (is_king_in_check_for_squares(
                            dest,
                            bishop_squares,
                            nb_bishop_squares))
                    {
                        return TRUE;
                    }
                }

                if (type == TOUR
                    || type == DAME
                    || type == ROI)
                {
                    if (is_king_in_check_for_squares(
                            dest,
                            rook_squares,
                            nb_rook_squares))
                    {
                        return TRUE;
                    }
                }

                if (type == CAVALIER)
                {
                    if (is_king_in_check_for_squares(
                            dest,
                            knight_squares,
                            nb_knight_squares))
                    {
                        return TRUE;
                    }
                }
            }

            if (type == PION
                && p->destinations[j].capturedPiece == PAS_DE_CAPTURE)
            {
                unsigned int k;
                for (k = 0; k < nb_pawn_promotion_squares; k++)
                {
                    if (dest == pawn_promotion_squares[k])
                    {
                        return TRUE;
                    }
                }
            }

            if (type == PION && dest == p->casePromotion)
            {
                type = p->typePromotion;
            }
            else j++;
        }
    }

    return FALSE;
}

static bool_t
is_in_check_colour(
    const piece_t* king,
    square_t* bishop_squares,
    square_t* rook_squares,
    square_t* knight_squares,
    square_t* pawn_promotion_squares)
{
    bool_t is_in_check = FALSE;
    if (king->caseInitiale == king->caseArrivee)
    {
        is_in_check = is_king_in_check(
            king->camp,
            bishop_squares,
            2,
            rook_squares,
            3,
            knight_squares,
            4,
            pawn_promotion_squares,
            2);
    }
    else if (king->castling == KING_SIDE)
    {
        is_in_check = is_king_in_check(
            king->camp,
            &bishop_squares[1],
            1,
            &rook_squares[2],
            1,
            &knight_squares[2],
            1,
            &pawn_promotion_squares[1],
            1);
    }
    else if (king->castling == QUEEN_SIDE)
    {
        is_in_check = is_king_in_check(
            king->camp,
            bishop_squares,
            1,
            rook_squares,
            1,
            knight_squares,
            1,
            pawn_promotion_squares,
            1);
    }
    return is_in_check;
}

bool_t
are_immobile_kings_checked(
    const piece_t* white_king,
    const piece_t* black_king,
    bool_t* must_white_king_switchback,
    bool_t* must_black_king_switchback)
{
    static square_t white_squares_bishop[] = {d2, f2};
    static square_t white_squares_rook[]   = {d1, e2, f1};
    static square_t white_squares_knight[] = {d3, c2, f3, g2};
    static square_t white_squares_pawn[]   = {d1, f1};

    static square_t black_squares_bishop[] = {d7, f7};
    static square_t black_squares_rook[]   = {d8, e7, f8};
    static square_t black_squares_knight[] = {d6, c7, f6, g7};
    static square_t black_squares_pawn[]   = {d8, f8};

    *must_white_king_switchback = FALSE;
    *must_black_king_switchback = FALSE;

    if (white_king->nbDestination == 1
        && is_in_check_colour(white_king,
                              white_squares_bishop,
                              white_squares_rook,
                              white_squares_knight,
                              white_squares_pawn))
    {
        if (NbCoupsBlancsRestants < 2) return TRUE;
        else *must_white_king_switchback = TRUE;
    }

    if (black_king->nbDestination == 1
        && is_in_check_colour(black_king,
                              black_squares_bishop,
                              black_squares_rook,
                              black_squares_knight,
                              black_squares_pawn))
    {
        if (NbCoupsNoirsRestants < 2) return TRUE;
        else *must_black_king_switchback = TRUE;
    }

    return FALSE;
}
