/*
 * test_moves_order.c --- Order the moves of the current position.
 *
 * Copyright (c) 2001-2004, 2009, 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include        "common.h"

#include	"test_moves_order.h"
#include	"test_moves_order_pawns.h"
#include	"test_moves_order_tools.h"
#include	"main.h"
#include	"distance.h"
#include	"pcpj.h"
#include	"twin_squares.h"
#include	"piece.h"
#include        "moves_order.h"

#include	<stdlib.h>	/* for exit(...) */
#include	<assert.h>

/*--------------------------------------------------------------------------*/

/* #define	DEBUG_TEST_MOVE_ORDER */
#include	"pcpjtool.h"

/*--------------------------------------------------------------------------*/

typedef enum
{
    e_Initial_Square,
    e_Destination_Square,
    e_Captured_Without_Moving
} constraint_type_t ;

/*--------------------------------------------------------------------------*/

static bool_t add_capture_constraints(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration );

static bool_t add_square_constraints(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration );

static bool_t add_castling_constraints(
    const piece_t*		king,
    const piece_t*		rook,
    struct moves_order_t*	moves_order ,
    const exploration_t*	exploration );

static bool_t add_pawns_on_same_column_constraints(
    struct moves_order_t* moves_order,
    const exploration_t*  exploration);

/*--------------------------------------------------------------------------*/

static bool_t add_constraint(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration,
    square_t			square,
    piece_index_t		piece_index,
    constraint_type_t		constraint_type,
    const char*                 constraint_type_name);

static void add_square_to_list( square_t	square,
                                square_t*	squares,
                                unsigned int*	nb_squares );

/*--------------------------------------------------------------------------*/

#define	MAX_DESTINATION_PER_TYPE	128

/*--------------------------------------------------------------------------*/

static void
find_squares_used_by_pieces(
    const exploration_t* exploration,
    const piece_t** wK,
    const piece_t** bK,
    const piece_t** wR,
    const piece_t** bR,
    square_t* king_squares,
    unsigned int* nb_king_squares,
    square_t* pawn_squares,
    unsigned int* nb_pawn_squares,
    square_t* queen_squares,
    unsigned int* nb_queen_squares,
    square_t* rook_squares,
    unsigned int* nb_rook_squares,
    square_t* bishop_squares,
    unsigned int* nb_bishop_squares,
    square_t* knight_squares,
    unsigned int* nb_knight_squares)
{
    piece_index_t i;

    *nb_king_squares = 0;
    *nb_pawn_squares = 0;
    *nb_queen_squares = 0;
    *nb_rook_squares = 0;
    *nb_bishop_squares = 0;
    *nb_knight_squares = 0;

    for (i = INDEX_ROI_BLANC; i <= INDEX_PION_NOIR_H; i++)
    {
        const piece_t*	piece = &exploration->pieces[i];
        unsigned int	j ;
        piece_type_t	piece_type = piece->typePiece ;

        if (has_castled(piece))
        {
            if ( piece_type == ROI )
            {
                if ( piece->camp == BLANC )
                {
                    *wK = piece;
                }
                else
                {
                    *bK = piece;
                }
            }
            else
            {
                if ( piece->camp == BLANC )
                {
                    *wR = piece;
                }
                else
                {
                    *bR = piece;
                }
            }
        }

        for (j = 0; j < piece->nbDestination; j++)
        {
            square_t square;
            if (j == 0)
            {
                if (piece->typePiece == ROI && has_castled(piece))
                {
                    continue;
                }

                square = piece->caseInitiale;
            }
            else
            {
                square = piece->destinations[j-1].destination;
            }

            if (square == CASE_ARRIVEE_QUELCONQUE)
            {
                continue;
            }

            if (square == piece->casePromotion)
            {
                piece_type = piece->typePromotion;
            }

            switch(piece_type)
            {
            case ROI:
                add_square_to_list(square, king_squares, nb_king_squares);
                break ;
            case PION:
                add_square_to_list(square, pawn_squares, nb_pawn_squares);
                break ;
            case DAME:
                add_square_to_list(square, queen_squares, nb_queen_squares);
                break ;
            case TOUR:
                add_square_to_list(square, rook_squares, nb_rook_squares);
                break ;
            case FOU:
                add_square_to_list(square,
                                   bishop_squares,
                                   nb_bishop_squares);
                break ;
            case CAVALIER:
                add_square_to_list(square,
                                   knight_squares,
                                   nb_knight_squares);
                break ;
            default:
                assert(FALSE);
                exit(EXIT_FAILURE);
            }
        }
    }
}

/*--------------------------------------------------------------------------*/

static bool_t
is_something_happening_on_square(
    const exploration_t* exploration, square_t square)
{
    piece_index_t i;

    for (i = INDEX_ROI_BLANC; i <= INDEX_PION_NOIR_H ; i++)
    {
        const piece_t* piece = &exploration->pieces[i];

        if (piece->typePiece == TOUR && has_castled(piece))
        {
            continue;
        }

        /* Look if pieces on their original square block other pieces.  */
        if (square == piece->caseInitiale &&
            piece->destinations[0].distance > 0)
        {
            return TRUE;
        }

        /* Piece captured without having moved may block other pieces.  */
        if (square == piece->caseInitiale
            && is_captured(piece)
            && piece->destinations[0].distance == 0)
        {
            return TRUE;
        }

        /* Look if piece that have reached their destination square block
         * other pieces.  */
        if (square == piece->caseArrivee
            && piece->destinations[0].distance > 0
            && !is_captured(piece)
            && ((piece->camp == BLANC && NbCoupsBlancsRestants < 2)
                || (piece->camp == NOIR && NbCoupsNoirsRestants < 2)
                || (!has_captured(piece, piece->nbDestination - 1)
                    && !(has_castled(piece)
                         /* && piece->typePiece == ROI : already done */
                         && piece->nbDestination == 1)
                    /* and not just promoted */
                    && !( square == piece->casePromotion
                          && ( promotion_index(piece)
                               == piece->nbDestination - 1)))))
        {
            return TRUE;
        }
    }
    return FALSE;
}

/*--------------------------------------------------------------------------*/

static bool_t
add_constraint_for_forbidden_squares(
    const exploration_t* exploration,
    struct moves_order_t* moves_order,
    twin_squares_t* twin_squares,
    square_t* king_squares,
    unsigned int nb_king_squares,
    square_t* pawn_squares,
    unsigned int nb_pawn_squares,
    square_t* queen_squares,
    unsigned int nb_queen_squares,
    square_t* rook_squares,
    unsigned int nb_rook_squares,
    square_t* bishop_squares,
    unsigned int nb_bishop_squares,
    square_t* knight_squares,
    unsigned int nb_knight_squares)
{
    square_t square;
    bool_t ok;

    for (square = a1 ; square <= h8 ; square++)
    {
        distance_t distance_table[6 * SIZE_PIECE];
        piece_index_t i;

        if (CasesInterdites[square]
            || !is_something_happening_on_square(exploration, square))
        {
            continue;
        }

        CasesInterdites[square] = TRUE;

        distance_create_table_for_squares(
            distance_table,
            king_squares,
            nb_king_squares,
            pawn_squares,
            nb_pawn_squares,
            queen_squares,
            nb_queen_squares,
            rook_squares,
            nb_rook_squares,
            bishop_squares,
            nb_bishop_squares,
            knight_squares,
            nb_knight_squares,
            twin_squares);

        distance_push(distance_table);

        for (i = INDEX_ROI_BLANC; i <= INDEX_PION_NOIR_H; i++)
        {
            const piece_t* piece = &exploration->pieces[i];

            if (piece->typePiece == TOUR && has_castled(piece))
            {
                continue;
            }

            /* Look if pieces on their original square block other pieces. */
            if (square == piece->caseInitiale &&
                piece->destinations[0].distance > 0)
            {
                ok = add_constraint(moves_order,
                                    exploration,
                                    square,
                                    i,
                                    e_Initial_Square,
                                    "e_Initial_Square");
                if (!ok) break;
            }

            /* Piece captured without having moved may block other pieces. */
            if (square == piece->caseInitiale
                && is_captured(piece)
                && piece->destinations[0].distance == 0)
            {
                ok = add_constraint(moves_order,
                                    exploration,
                                    square,
                                    i,
                                    e_Captured_Without_Moving,
                                    "e_Captured_Without_Moving");
                if (!ok) break;
            }

            /* Look if piece that have reached their destination square block
             * other pieces.  */
            if (square == piece->caseArrivee
                && piece->destinations[0].distance > 0
                && !is_captured(piece)
                && ((piece->camp == BLANC && NbCoupsBlancsRestants < 2)
                    || (piece->camp == NOIR && NbCoupsNoirsRestants < 2)
                    || (!has_captured(piece, piece->nbDestination - 1)
                        && !(has_castled(piece)
                             /* && piece->typePiece == ROI : already done */
                             && piece->nbDestination == 1 )
                        /* and not just promoted */
                        && !(square == piece->casePromotion
                             && (promotion_index(piece)
                                 == piece->nbDestination - 1)))))
            {
                ok = add_constraint(moves_order,
                                    exploration,
                                    square,
                                    i,
                                    e_Destination_Square,
                                    "e_Destination_Square");
                if (!ok) break;
            }
        }

        distance_pop();
        CasesInterdites[ square ] = FALSE;

        if (!ok) return FALSE;
    }
    return TRUE;
}

/*--------------------------------------------------------------------------*/

bool_t
test_moves_order(exploration_t* exploration)
{
    struct moves_order_t*	moves_order = create_moves_order( exploration );
    const piece_t*		wk = NULL ;
    const piece_t*		bk = NULL ;
    const piece_t*		wr = NULL ;
    const piece_t*		br = NULL ;
    twin_squares_t twin_squares;

    square_t king_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_king_squares;
    square_t pawn_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_pawn_squares;
    square_t queen_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_queen_squares;
    square_t rook_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_rook_squares;
    square_t bishop_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_bishop_squares;
    square_t knight_squares[MAX_DESTINATION_PER_TYPE];
    unsigned int nb_knight_squares;

    if (!add_capture_constraints(moves_order, exploration)
        || !add_square_constraints(moves_order, exploration)
        || !add_pawns_on_same_column_constraints(moves_order, exploration))
    {
        return FALSE;
    }

    find_squares_used_by_pieces(exploration,
                                &wk, &bk, &wr, &br,
                                king_squares, &nb_king_squares,
                                pawn_squares, &nb_pawn_squares,
                                queen_squares, &nb_queen_squares,
                                rook_squares, &nb_rook_squares,
                                bishop_squares, &nb_bishop_squares,
                                knight_squares, &nb_knight_squares);

    if ((wk != NULL
         && !add_castling_constraints(wk, wr, moves_order, exploration))
        || (bk != NULL
            && !add_castling_constraints(bk, br, moves_order, exploration)))
    {
        return FALSE;
    }

    /* initialisations for distance computation */
    create_twin_squares(&exploration->pieces[INDEX_ROI_BLANC],
                        exploration->nb_moving_pieces,
                        &twin_squares);

    if (!add_constraint_for_forbidden_squares(
        exploration,
        moves_order,
        &twin_squares,
        king_squares, nb_king_squares,
        pawn_squares, nb_pawn_squares,
        queen_squares, nb_queen_squares,
        rook_squares, nb_rook_squares,
        bishop_squares, nb_bishop_squares,
        knight_squares, nb_knight_squares))
    {
        return FALSE;
    }

    if (!add_pawn_line_constraints(moves_order, exploration, &twin_squares))
    {
        return FALSE;
    }

    remove_weak_constraints(moves_order, exploration);
    if (MainPrintPositionLevel > 1)
    {
        print_moves_order( MainFD, moves_order, exploration );
    }

    create_predecessors(moves_order, exploration);

    return TRUE;
}

/*--------------------------------------------------------------------------*/

static bool_t add_capture_constraints(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration )
{
    piece_index_t	i ;

    /* Captured pieces must reach there destination before the capture. */
    for ( i = INDEX_ROI_BLANC ; i <= INDEX_PION_NOIR_H ; i++ )
    {
        const piece_t*	piece = &exploration->pieces[ i ];
        unsigned int	j ;

        for ( j = 0 ; j < piece->nbDestination ; j++ )
        {
            piece_index_t index = piece->destinations[j].capturedPiece;
            if ( index != PAS_DE_CAPTURE )
            {
                const piece_t* captured_piece = &exploration->pieces[ index ];
                if (captured_piece->destinations[0].distance > 0)
                {
                    /* Captured piece has moved */
                    if ( !add_one_order(
                             moves_order,
                             index,
                             captured_piece->nbDestination - 1,
                             e_Last_Move,
                             i,
                             j,
                             e_Last_Move,
                             exploration,
                             "captured piece must reach destination" ) )
                    {
                        return FALSE ;
                    }
                }
            }
        }
    }
    return TRUE ;
}

/*--------------------------------------------------------------------------*/

static bool_t add_square_constraints(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration )
{
    square_t		square ;
    const piece_t*	wK = &exploration->pieces[ exploration->board[ e1 ] ];
    const piece_t*	bK = &exploration->pieces[ exploration->board[ e8 ] ];

    for ( square = a1 ; square <= h8 ; square++ )
    {
        const piece_t*	orig = NULL ;
        const piece_t*	dest = NULL ;
        piece_index_t	i ;

        /* This is only used to check the evacuation of a castling rook.  The
         * other cases are handled by add_constraint for e_Initial_Square. */
        if ( square == h1 && wK->castling == KING_SIDE )
        {
            orig = wK ;
        }
        else if ( square == a1 && wK->castling == QUEEN_SIDE )
        {
            orig = wK ;
        }
        else if ( square == h8 && bK->castling == KING_SIDE )
        {
            orig = bK ;
        }
        else if ( square == a8 && bK->castling == QUEEN_SIDE )
        {
            orig = bK ;
        }

        for ( i = INDEX_ROI_BLANC ;
              dest == NULL && i <= INDEX_PION_NOIR_H ;
              i++ )
        {
            const piece_t* piece = &exploration->pieces[ i ];

            if ((piece->destinations[piece->nbDestination - 1].destination
                 == square)
                && piece->destinations[0].distance != 0
                && !is_captured(piece)
                /* If the piece made a capture (or promoted) on the last move
                 * and may swithback, then we don't make any assumption on the
                 * move order.  See tests `promotee_must_switchback2.test' for
                 * capture, `promotee_must_switchback3.test' for promotion, and
                 * also `promotee_must_switchback4.test'.
                 */
                && !((has_captured(piece, piece->nbDestination - 1)
                      || (square == piece->casePromotion
                          && (promotion_index(piece) == piece->nbDestination-1)))
                     && ((piece->camp == BLANC && NbCoupsBlancsRestants > 1)
                         || (piece->camp == NOIR && NbCoupsNoirsRestants > 1))))
            {
                dest = piece;
            }
        }

        if ( orig == NULL && dest == NULL )
        {
            continue ;
        }

        for ( i = INDEX_ROI_BLANC ; i <= INDEX_PION_NOIR_H ; i++ )
        {
            const piece_t*	piece = &exploration->pieces[ i ];
            int			move_index ;

            if (piece->destinations[0].distance == 0)
            {
                continue ;
            }

            if ( orig != NULL && piece != orig )
            {
                /* This is only used to check the evacuation of a castling rook.
                 * The other cases are handled by add_constraint for
                 * e_Initial_Square. */
                for ( move_index = 0 ;
                      move_index < piece->nbDestination ;
                      move_index++ )
                {
                    if ( piece->destinations[move_index].destination == square)
                    {
                        if ( !add_one_order( moves_order,
                                             orig->index,
                                             0,
                                             e_First_Move,
                                             piece->index,
                                             move_index,
                                             e_Last_Move,
                                             exploration,
                                             "castling rook must move" ) )
                        {
                            return FALSE ;
                        }
                    }
                }
            }

            if ( dest != NULL
                 && piece != dest
                 && piece->nbDestination >= 2 )
            {
                for ( move_index = piece->nbDestination - 2 ;
                      move_index >= 0 ;
                      move_index-- )
                {
                    if (piece->destinations[move_index].destination == square)
                    {
                        if ( !add_one_order( moves_order,
                                             piece->index,
                                             move_index + 1,
                                             e_First_Move,
                                             dest->index,
                                             dest->nbDestination - 1,
                                             e_Last_Move,
                                             exploration,
                                             "pass before final piece arrives"))
                        {
                            return FALSE ;
                        }
                    }
                }
            }
        }
    }

    return TRUE ;
}

/*--------------------------------------------------------------------------*/

static bool_t add_castling_constraints(
    const piece_t*		king,
    const piece_t*		rook,
    struct moves_order_t*	moves_order ,
    const exploration_t*	exploration )
{
    square_t	intermediates[ 3 ];
    bool_t	queen_side ;
    int		i ;
    square_t	destSquare;
    const destination_t* destination;
    bool_t      can_make_switchback;

    /* Castling occurs before first move of the rook */
    if ( rook->destinations[0].distance != 0)
    {
        if ( !add_one_order( moves_order,
                             king->index,
                             0,
                             e_Last_Move,
                             rook->index,
                             0,
                             e_First_Move,
                             exploration,
                             "castling : occurs before 1st move of rook" ) )
        {
            return FALSE ;
        }
    }
    else if (is_captured(rook))
    {
        const piece_t* piece = &exploration->pieces[rook->pieceCapturante];
        for (i = 0; i < piece->nbDestination; i++)
        {
            if (piece->destinations[i].capturedPiece == rook->index)
                break;
        }
        if (!add_one_order(
                moves_order,
                king->index,
                0,
                e_Last_Move,
                piece->index,
                i,
                e_Last_Move,
                exploration,
                "castling : occurs before rook is captured without moving" ) )
        {
            return FALSE;
        }
    }

    if ( king->castling == KING_SIDE )
    {
        intermediates[ 0 ] = king->caseInitiale + 1 ;
        intermediates[ 1 ] = king->caseInitiale + 2 ;
        intermediates[ 2 ] = CASE_ARRIVEE_QUELCONQUE;
        queen_side = FALSE ;
    }
    else
    {
        intermediates[ 0 ] = king->caseInitiale - 1 ;
        intermediates[ 1 ] = king->caseInitiale - 2 ;
        intermediates[ 2 ] = king->caseInitiale - 3 ;
        queen_side = TRUE ;
    }

    for ( i = INDEX_ROI_BLANC ; i <= INDEX_PION_NOIR_H ; i++ )
    {
        const piece_t*	piece = &exploration->pieces[ i ];

        /* Castling occurs after intermediate squares have been emptied */
        if ( piece != rook
             && ( piece->caseInitiale == intermediates[ 0 ]
                  || piece->caseInitiale == intermediates[ 1 ]
                  || ( queen_side
                       && piece->caseInitiale == intermediates[ 2 ] ) ) )
        {
            bool_t	ok ;

            if ( piece->destinations[0].distance == 0
                 && NbCoupsBlancsRestants < 2
                 && NbCoupsNoirsRestants  < 2 )
            {
                bool_t		captured_on_square = TRUE;
                piece_index_t	capturer_index 	   = piece->pieceCapturante;
                piece_index_t	captured_index     = piece->index;
                int		nbCaptures 	   = 0;
                int		move_index ;
                while ( captured_on_square )
                {
                    const piece_t* capturer
                        = &exploration->pieces[ capturer_index ];

                    nbCaptures++;

                    move_index = 0 ;
                    while (capturer->destinations[move_index].capturedPiece
                           != captured_index)
                    {
                        move_index++;
                    }
                    if ( move_index == capturer->nbDestination - 1 )
                    {
                        if (!is_captured(capturer))
                        {
#ifdef DEBUG_TEST_MOVE_ORDER
                            fprintf( MainFD,
                                     "Impossible move order.  See file %s:%d\n",
                                     __FILE__, __LINE__ );
#endif /* DEBUG_TEST_MOVE_ORDER */
                            return FALSE ;
                        }
                        capturer_index = capturer->pieceCapturante ;
                        captured_index = capturer->index ;
                    }
                    else
                    {
                        captured_on_square = FALSE;
                    }
                }
                if ( nbCaptures == 1 )
                {
                    ok = add_one_order(
                        moves_order,
                        capturer_index,
                        move_index + 1,
                        e_First_Move,
                        king->index,
                        0,
                        e_First_Move,
                        exploration,
                        "castling : blocking piece must be captured" );
                }
                else
                {
                    ok = TRUE;
                }
            }
            else if (piece->destinations[0].distance > 0)
            {
                ok = add_one_order(
                    moves_order,
                    piece->index,
                    0,
                    e_First_Move,
                    king->index,
                    0,
                    e_First_Move,
                    exploration,
                    "castling : blocking piece must move before castling" );
            }
            else
            {
                ok = TRUE;
            }
            if ( !ok )
            {
                return FALSE;
            }
        }

        /* Castling occurs before intermediate squares are definitively
         * occupied */
        destination = &piece->destinations[piece->nbDestination - 1];
        destSquare = destination->destination;

        can_make_switchback = TRUE;
        if (destination->capturedPiece != PAS_DE_CAPTURE
            && ((piece->camp == BLANC && exploration->white_spare_moves < 2)
                || (piece->camp == NOIR && exploration->black_spare_moves < 2)))
        {
            can_make_switchback = FALSE;
        }

        if ( piece != king
             && piece != rook
             && !is_captured(piece)
             && ( destSquare == intermediates[ 0 ]
                  || destSquare == intermediates[ 1 ]
                  || ( queen_side && destSquare == intermediates[ 2 ] ) )
             && !can_make_switchback )
        {
            if ( !add_one_order(
                     moves_order,
                     king->index,
                     0,
                     e_Last_Move,
                     piece->index,
                     piece->nbDestination - 1,
                     e_Last_Move,
                     exploration,
        "castling : intermediate square cannot be definitively occupied" ) )
            {
                return FALSE ;
            }
        }
    }

    /* Castling occures before origin squares of King or Rook are needed */
    /* This is done for the King but not for the Rook in the general algotithm.
     */

    return TRUE ;
}

/*--------------------------------------------------------------------------*/

static bool_t
add_constraint_for_pawn_on_same_column(
    const piece_t* before_piece,
    const piece_t* after_piece,
    struct moves_order_t* moves_order,
    const exploration_t*  exploration)
{
    piece_index_t before_piece_index = INDEX_CASE_VIDE;
    unsigned int before_move_index = 0;
    move_type_t before_type;
    char* description;

    if (has_captured(before_piece, 0))
    {
        /* This must be checked first, to handle the case of a pawn that
         * captures first, and then moves.
         */
        before_piece_index = before_piece->index;
        before_move_index = 0;
        before_type = e_First_And_Last_Move;
        description = "pawn captures before other pawn passes";
    }
    else if (before_piece->nbDestination > 1)
    {
        before_piece_index = before_piece->index;
        before_move_index = 1;
        before_type = e_First_And_Last_Move;
        description = "pawn moves and captures before other pawn passes";
    }
    else
    {
        find_move_after_capture(
            before_piece->index,
            exploration,
            &before_piece_index,
            &before_move_index);
        before_type = e_First_Move;
        description = "pawn is captured before other pawn passes";
    }

    return add_one_order(
        moves_order,
        before_piece_index,
        before_move_index,
        before_type,
        after_piece->index,
        0,
        e_Last_Move,
        exploration,
        description);
}

static bool_t
add_pawns_on_same_column_constraints(
    struct moves_order_t* moves_order,
    const exploration_t*  exploration)
{
    piece_index_t white_pawns[8], black_pawns[8];
    unsigned int i;

    find_pawns(exploration, white_pawns, black_pawns);

    for (i = 0; i < 8; i++)
    {
        const piece_t* white_piece = &exploration->pieces[white_pawns[i]];
        const piece_t* black_piece = &exploration->pieces[black_pawns[i]];

        if (row(white_piece->destinations[0].destination) >
            row(black_piece->destinations[0].destination))
        {
            if ((white_piece->nbDestination == 1
                 && !is_captured(white_piece)
                 && !has_captured(white_piece, 0))
                || white_piece->destinations[0].distance >= 4)
            {
                if (!add_constraint_for_pawn_on_same_column(
                        black_piece,
                        white_piece,
                        moves_order,
                        exploration))
                    return FALSE;
            }
            else if ((black_piece->nbDestination == 1
                      && !is_captured(black_piece)
                      && !has_captured(black_piece, 0))
                     || black_piece->destinations[0].distance >= 4)
            {
                if (!add_constraint_for_pawn_on_same_column(
                        white_piece,
                        black_piece,
                        moves_order,
                        exploration))
                    return FALSE;
            }
        }
    }
    return TRUE;
}

/*--------------------------------------------------------------------------*/

static void
add_square_to_list(
    square_t		square,
    square_t*		squares,
    unsigned int*	nb_squares )
{
    unsigned int	i ;

    assert( *nb_squares < MAX_DESTINATION_PER_TYPE );

    for ( i = 0 ; i < *nb_squares ; i++ )
    {
        if ( squares[ i ] == square )
        {
            break ;
        }
    }

    if ( i == *nb_squares )
    {
        squares[ i ] = square ;
        (*nb_squares)++ ;
    }
}

/*--------------------------------------------------------------------------*/

static distance_t delta_pawn(square_t square, square_t from, square_t to)
{
    square_t min;
    square_t max;
    if (from < to)
    {
        min = from;
        max = to;
    }
    else
    {
        min = to;
        max = from;
    }
    while (min <= max)
    {
        if (min == square)
        {
            return  INFINI;
        }
        min += 16;
    }
    return 0;
}

/*--------------------------------------------------------------------------*/

static bool_t
add_constraint(
    struct moves_order_t*	moves_order,
    const exploration_t*	exploration,
    square_t			square,
    piece_index_t		piece_index,
    constraint_type_t		constraint_type,
    const char*                 constraint_type_name)
{
    unsigned int	i ;
    const piece_t*	p ;

    /* Used if captured_without_moving : we initialise them to avoid a compiler
     * warning. */
    piece_index_t capturing_piece_index = 0;
    unsigned int capturing_move_index = 0;

    char description[64];

    if ( constraint_type == e_Captured_Without_Moving )
    {
        find_move_after_capture(
            piece_index,
            exploration,
            &capturing_piece_index,
            &capturing_move_index);
    }

    for ( i = 0, p = &( exploration->pieces[ INDEX_ROI_BLANC ] ) ;
          i < exploration->nb_moving_pieces ;
          i++, p++ )
    {
        piece_type_t piece_type = p->typePiece;

        /* Skip the piece that must not be tested */
        if (p->destinations[0].distance == 0)
        {
            continue ;
        }
        if ( constraint_type == e_Initial_Square
             && p->caseInitiale == square )
        {
            continue ;
        }
        if ( constraint_type == e_Destination_Square &&
             i + INDEX_ROI_BLANC == piece_index )
        {
            continue ;
        }

        unsigned int nb_spare_moves =
            p->camp == BLANC ? NbCoupsBlancsRestants : NbCoupsNoirsRestants;

        if ( constraint_type == e_Destination_Square )
        {
            int move_index;
            int index_promotion = p->nbDestination;

            if (has_promoted(p))
            {
                piece_type = p->typePromotion ;
                index_promotion = promotion_index( p );
            }

            for ( move_index = p->nbDestination - 1 ;
                  move_index >= 0 ;
                  move_index-- )
            {
                square_t current_square;

                if (move_index == 0 && p->typePiece == ROI && has_castled(p))
                {
                    continue ;
                }

                if (is_captured(p)
                    && move_index == p->nbDestination - 1
                    && square == p->destinations[move_index].destination)
                {
                    /* Already done by add_capture_constraints */
                    continue;
                }

                if ( move_index == 0 )
                {
                    current_square = p->caseInitiale ;
                }
                else
                {
                    current_square = p->destinations[move_index - 1].destination;
                }

                if (current_square == square)
                {
                    /* Check already done in add_square_constraints */
                    break;
                }

                if ( move_index == index_promotion )
                {
                    piece_type = PION ;
                }

                if ( current_square != CASE_ARRIVEE_QUELCONQUE
                     && p->destinations[move_index].destination !=
                     CASE_ARRIVEE_QUELCONQUE )
                {
                    distance_t delta;
                    if (piece_type == PION && (!has_captured(p, move_index)))
                    {
                        delta = delta_pawn(
                            square,
                            current_square,
                            p->destinations[move_index].destination);
                    }
                    else
                    {
                        delta = distance_delta(
                            piece_type,
                            p->camp,
                            current_square,
                            p->destinations[move_index].destination);
                    }
                    if (delta > nb_spare_moves)
                    {
                        if ( !add_one_order(
                                 moves_order,
                                 i + INDEX_ROI_BLANC,
                                 move_index,
                                 e_First_Move,
                                 piece_index,
                                 exploration->pieces[
                                     piece_index ].nbDestination - 1,
                                 e_Last_Move,
                                 exploration,
                                 constraint_type_name))
                        {
                            return FALSE ;
                        }
                        break ;
                    }
                }
            }
        }
        else
        {
            unsigned int move_index;
            square_t current_square = p->caseInitiale;

            for ( move_index = 0 ;
                  move_index < p->nbDestination ;
                  move_index++ )
            {
                const destination_t* dest;

                if (move_index == 0 && p->typePiece == ROI && has_castled(p))
                {
                    current_square = p->destinations[0].destination;
                    continue ;
                }

                dest = &p->destinations[move_index];

                /* The other constraints are added because captured pieces must
                 * reach their destination square.  So the following test in
                 * enough.
                 */
                if (constraint_type == e_Captured_Without_Moving
                    && dest->destination == square
                    && dest->capturedPiece != PAS_DE_CAPTURE)
                {
                    break;
                }

                if ( current_square == p->casePromotion )
                {
                    /* If the piece return to its promotion square, this is
                     * still correct. */
                    piece_type = p->typePromotion ;
                }

                if ( current_square != CASE_ARRIVEE_QUELCONQUE
                     && current_square != square
                     && dest->destination != CASE_ARRIVEE_QUELCONQUE )
                {
                    distance_t delta;
                    if (piece_type == PION && !has_captured(p, move_index))
                    {
                        delta = delta_pawn(
                            square, current_square, dest->destination);
                    }
                    else
                    {
                        delta = distance_delta(
                            piece_type,
                            p->camp,
                            current_square,
                            dest->destination);
                    }

                    if (delta > nb_spare_moves)
                    {
                        piece_index_t	before_piece_index;
                        unsigned int	before_move_index;
                        move_type_t	before_type;

                        switch( constraint_type )
                        {
                        case e_Captured_Without_Moving:
                        {
                            before_piece_index = capturing_piece_index;
                            before_move_index = capturing_move_index;
                            before_type = e_First_Move;
                        }
                        break;
                        case e_Initial_Square:
                        {
                            before_piece_index = piece_index ;
                            before_move_index  = 0 ;
                            before_type        = e_First_Move ;
                        }
                        break;
                        default:
                            assert( FALSE );
                            exit( 1 );
                            break;
                        }

                        sprintf(description, "2 : %s",
                                constraint_type_name);
                        if ( !add_one_order(
                                 moves_order,
                                 before_piece_index,
                                 before_move_index,
                                 before_type,
                                 i + INDEX_ROI_BLANC,
                                 move_index,
                                 e_Last_Move,
                                 exploration,
                                 description))
                        {
                            return FALSE ;
                        }
                        break ;
                    }
                }
                current_square = dest->destination;
            }
        }
    }

    return TRUE ;
}
