/*
 * trajet.c --- verify trajectories.
 *
 * Copyright (c) 1997-2002, 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"common.h"

#include	"trajet.h"

#include	"pcpj.h"
#include	"distance.h"
#include	"main.h"
#include        "immobile_kings_in_check.h"
#include	"precise_trajectories.h"
#include        "pcpjtool.h"

/*--------------------------------------------------------------------------*/

static bool_t are_pawn_trajectories_correct(const twin_squares_t* twin_squares);
static bool_t are_pawns_on_same_column_colliding(void);
static bool_t is_pawn_trajectory_conflicting_with_twin_square(
    const twin_squares_t* twin_squares,
    square_t pawn_origin,
    square_t current,
    square_t dest,
    piece_index_t capturedPiece,
    int deltaY);
static void add_king_switchback(piece_t* king);
static void remove_king_switchback(piece_t* king);
static void update_trajectories_due_to_stationary_pieces(
    const twin_squares_t* twin_squares);

/*--------------------------------------------------------------------------*/

void
trajet(const twin_squares_t* twin_squares)
{
    piece_t* white_king = &ListeSures[0];
    piece_t* black_king = &ListeSures[1];
    bool_t must_white_king_switchback;
    bool_t must_black_king_switchback;
    twin_squares_t twin_squares_after_king_swithback;
    const twin_squares_t* twin_squares_to_use = twin_squares;

    /* Checking for immobile kings must be done first, as it may use spare
       moves. */
    if (are_immobile_kings_checked(
            white_king, black_king,
            &must_white_king_switchback, &must_black_king_switchback))
    {
        return;
    }

    if (must_white_king_switchback)
    {
        add_king_switchback(white_king);
        NbCoupsBlancsRestants -= 2;
    }
    if (must_black_king_switchback)
    {
        add_king_switchback(black_king);
        NbCoupsNoirsRestants -= 2;
    }

    if (must_white_king_switchback || must_black_king_switchback)
    {
        create_twin_squares(ListeSures,
                            NbSures + NbRestantes,
                            &twin_squares_after_king_swithback);
        twin_squares_to_use = &twin_squares_after_king_swithback;
    }

    if (are_pawn_trajectories_correct(twin_squares_to_use)
        && !are_pawns_on_same_column_colliding())
    {
        update_trajectories_due_to_stationary_pieces(twin_squares_to_use);
    }

    if (must_white_king_switchback)
    {
        remove_king_switchback(white_king);
        NbCoupsBlancsRestants += 2;
    }
    if (must_black_king_switchback)
    {
        remove_king_switchback(black_king);
        NbCoupsNoirsRestants += 2;
    }
}

/* Used by exploration for the --watch option. */
bool_t
trajet_traite_nouvelle_case_interdite(
    const exploration_t*	datas,
    square_t			nouvelle_case_interdite )
{
    unsigned int	i;
    const piece_t*	p ;
    distance_t		distance_table[ 6 * SIZE_PIECE ];
    twin_squares_t      twin_squares;

    CasesInterdites[ nouvelle_case_interdite ] = TRUE;

    create_twin_squares(ListeSures, NbSures + NbRestantes, &twin_squares);
    distance_create_table(distance_table, &twin_squares);
    distance_push( distance_table );

    for ( i = 0, p = &( datas->pieces[ INDEX_ROI_BLANC ] ) ;
          i < datas->nb_moving_pieces ;
          i++, p++ )
    {
        int		indice_destination;
        piece_type_t	type_piece = p->typePiece;
        square_t	position_actuelle =
            datas->pieces_squares[ i + INDEX_ROI_BLANC ];

        if (p->indDestAct == p->nbDestination - 1 &&
            position_actuelle == p->caseArrivee)
        {
            continue;
        }

        for (indice_destination = p->indDestAct;
             indice_destination < p->nbDestination;
             indice_destination++)
        {
            if (p->casePromotion != CASE_PAS_DE_PROMOTION &&
                position_actuelle == p->casePromotion)
            {
                /* Si la piece retourne sur sa case de promotion,
                 * cela ne pose aucun probleme. */
                type_piece = p->typePromotion;
            }

            if (position_actuelle != CASE_ARRIVEE_QUELCONQUE &&
                p->destinations[indice_destination].destination !=
                CASE_ARRIVEE_QUELCONQUE )
            {
                distance_t	delta = distance_delta(
                    type_piece,
                    p->camp,
                    position_actuelle,
                    p->destinations[indice_destination].destination);
                if (delta > 0)
                {
                    if (( p->camp == BLANC && delta > NbCoupsBlancsRestants ) ||
                        ( p->camp == NOIR  && delta > NbCoupsNoirsRestants  ) )
                    {
/*  fprintf(MainFD, */
/*  	"Position OUT a cause de la piece : 0x%02x. Type = %d.\n", */
/*  	p->caseInitiale, p->typePiece); */
/*  fprintf(MainFD, */
/*  	"position_actuelle = 0x%02x. destination = 0x%02x. " */
/*  	position_actuelle, p->destination[indice_destination]); */
/*  fprintf(MainFD, */
/*  	"distance avant = %d, apres = %d. Delta = %d\n", */
/*  	p->distance[indice_destination], */
/*  	distance(type_piece, p->camp, position_actuelle, */
/*  		 p->destination[indice_destination]), */
/*  	delta); */
                        distance_pop();
                        CasesInterdites[ nouvelle_case_interdite ] = FALSE ;
                        return FALSE ;
                    }
                }
            }
            position_actuelle = p->destinations[indice_destination].destination;
        }
    }

    distance_pop();
    CasesInterdites[ nouvelle_case_interdite ] = FALSE ;
    return TRUE;
}

/*--------------------------------------------------------------------------*/

static bool_t
are_pawn_trajectories_correct(const twin_squares_t* twin_squares)
{
    unsigned int	i, j;
    piece_t*		piece;
    square_t		initial_square;
    square_t		destination_square;
    column_t		depX, arrX;
    square_delta_t	deltaY;

    for (i=0, piece=ListeSures; i<NbSures+NbRestantes; i++, piece++)
    {
        if (piece->typePiece != PION)	continue;

        if (piece->nbDestination == 1 &&
            piece->caseInitiale == piece->destinations[0].destination)
        {
            continue;
        }

        if (piece->camp == BLANC)	deltaY = 16;
        else				deltaY = -16;

        initial_square = piece->caseInitiale;
        depX = column(initial_square);

        for (j=0; j<piece->nbDestination; j++)
        {
            square_t	verified_square;
            row_t	arrY;

            destination_square = piece->destinations[j].destination;
            arrX = column(destination_square);

            /* Verify that if a pawn captures, he moves to an adjacent column.*/
            if (piece->destinations[j].capturedPiece == PAS_DE_CAPTURE)
            {
                if (depX != column(destination_square))
                {
                    return FALSE;
                }
            }
            else
            {
                if (arrX - depX != 1 && depX - arrX != 1)
                {
                    return FALSE;
                }
            }

            if (is_pawn_trajectory_conflicting_with_twin_square(
                    twin_squares,
                    piece->caseInitiale,
                    initial_square,
                    destination_square,
                    piece->destinations[j].capturedPiece,
                    deltaY))
                return FALSE;

            /* Verify that a pawn does use a forbidden square. */
            arrY = row(destination_square);
            /* Peut etre considere comme inutile car la case d'arrivee
             * n'est jamais une case interdite. Si une prise y a lieu
             * alors la piece prise peut l'atteindre. Sinon c'est soit
             * la case d'arrivee du pion, soit sa case de promotion qui
             * ne peuvent pas etre interdite non plus.
             * En fait ce test sert pour les pions immobiles, ceux qui
             * se sont fait prendre sur place !?.
             */
            if (piece->destinations[j].capturedPiece == PAS_DE_CAPTURE)
                arrY += deltaY;

            verified_square = initial_square + deltaY;
            /* arrY may have the value 0xF0 (due to `arrY += deltaY') ! */
            while ((verified_square & 0xF0) != arrY)
            {
                if (CasesInterdites[verified_square])
                {
                    return FALSE;
                }
                verified_square += deltaY;
            }

            if (piece->casePromotion == destination_square)	break;

            initial_square = destination_square;
            depX = column(initial_square);
        }
    }

    return TRUE;
}

static bool_t
is_pawn_concerned_on_column(piece_t* p)
{
    return
        p->destinations[0].distance >= 4
        ||
        (p->nbDestination == 1
         && p->destinations[0].capturedPiece == PAS_DE_CAPTURE
         && p->pieceCapturante == PIECE_PAS_CAPTUREE);
}

static bool_t
are_pawns_on_same_column_colliding()
{
    piece_index_t index_white, index_black;

    for (index_white = INDEX_PION_BLANC_A, index_black = INDEX_PION_NOIR_A;
         index_white <= INDEX_PION_BLANC_H;
         index_white++, index_black++)
    {
        piece_t* white_piece = TabIndexPiece[index_white];
        piece_t* black_piece = TabIndexPiece[index_black];

        if (is_pawn_concerned_on_column(white_piece)
            && is_pawn_concerned_on_column(black_piece)
            && (row(white_piece->destinations[0].destination) >
                row(black_piece->destinations[0].destination)))
        {
            return TRUE;
        }
    }
    return FALSE;
}

static void
add_king_switchback(piece_t* king)
{
    unsigned int i = king->castling == NO_CASTLING ? 0 : 1;
    destination_t* dest = &king->destinations[i];
    dest->destination = CASE_ARRIVEE_QUELCONQUE;
    dest->capturedPiece = PAS_DE_CAPTURE;
    dest->distance = 1;
    dest->nb_predecessors_first = 0;
    dest->nb_predecessors_last = 0;
    dest->successors_first = NULL;
    dest->successors_last = NULL;
    dest->have_successors_decremented = FALSE;
    dest->is_successors_list_first_done = FALSE;
    dest->is_successors_list_last_done = FALSE;

    i++;
    dest = &king->destinations[i];
    dest->destination = king->caseArrivee;
    dest->capturedPiece = PAS_DE_CAPTURE;
    dest->distance = 1;
    dest->nb_predecessors_first = 0;
    dest->nb_predecessors_last = 0;
    dest->successors_first = NULL;
    dest->successors_last = NULL;
    dest->have_successors_decremented = FALSE;
    dest->is_successors_list_first_done = FALSE;
    dest->is_successors_list_last_done = FALSE;

    king->nbDestination = i + 1;
}

static void
remove_king_switchback(piece_t* king)
{
    if (king->castling == NO_CASTLING)
    {
        king->destinations[0].destination = king->caseArrivee;
        king->destinations[0].distance = 0;
    }
    king->nbDestination = 1;
}

static void
update_trajectories_due_to_stationary_pieces(const twin_squares_t* twin_squares)
{
    typedef struct
    {
        int		piece_indice;
        UCHAR		destination_index;
        distance_t	previous_distance;
    } new_distance_t;
    const unsigned int	max_nb_of_new_distances = 20;
    new_distance_t	new_distances[max_nb_of_new_distances];

    /* Cannot be > NbCoupsBlancsRestants + NbCoupsNoirsRestants */
    unsigned int	nb_new_distances = 0;

    piece_t*		piece;
    square_t		new_forbidden_squares[16];
    unsigned int	nb_new_forbidden_squares = 0;
    unsigned int	i, indice;
    int			white_spare_moves;
    int			black_spare_moves;
    distance_t		distance_table[6 * SIZE_PIECE];

    for (indice = 0, piece = &ListeSures[indice];
         indice < NbSures + NbRestantes;
         indice++, piece++)
    {
        if (piece->nbDestination   != 1                   ||
            piece->caseArrivee     != piece->caseInitiale ||
            piece->pieceCapturante != PIECE_PAS_CAPTUREE  )
        {
            continue;
        }

        if ( ( piece->typePiece == ROI || piece->typePiece == TOUR) &&
             ( piece->castling == KING_SIDE || piece->castling == QUEEN_SIDE ) )
        {
            continue;
        }

        if ((piece->camp == BLANC && NbCoupsBlancsRestants < 2) ||
            (piece->camp == NOIR  && NbCoupsNoirsRestants  < 2) )
        {
            CasesInterdites[piece->caseInitiale] = TRUE;
            new_forbidden_squares[nb_new_forbidden_squares] =
                piece->caseInitiale;
            nb_new_forbidden_squares++;
            continue;
        }
    }

    distance_create_table(distance_table, twin_squares);
    distance_push(distance_table);

    white_spare_moves = NbCoupsBlancsRestants;
    black_spare_moves = NbCoupsNoirsRestants;

    for (indice = 0, piece = &ListeSures[indice];
         indice < NbSures + NbRestantes;
         indice++, piece++)
    {
        int		destination_index;
        square_t	current_position = piece->caseInitiale;
        piece_type_t	type_piece       = piece->typePiece;

        if (piece->nbDestination == 1 &&
            piece->caseInitiale  == piece->caseArrivee)
        {
            continue;
        }

        for (destination_index = 0;
             destination_index < piece->nbDestination;
             destination_index++)
        {
            if (piece->casePromotion != CASE_PAS_DE_PROMOTION &&
                current_position     == piece->casePromotion)
            {
                /* Si la piece retourne sur sa case de promotion,
                 * cela ne pose aucun probleme. */
                type_piece = piece->typePromotion;
            }

            /* We don't make the test if :
             *	- the origin square is any square,
             *	- the destination square is any square,
             *	- a king is castling.
             */
            if ( current_position != CASE_ARRIVEE_QUELCONQUE &&
                 piece->destinations[destination_index].destination !=
                 CASE_ARRIVEE_QUELCONQUE &&
                ( piece->typePiece != ROI || piece->castling == NO_CASTLING ) )
            {
                distance_t new_distance = distance(
                    type_piece,
                    piece->camp,
                    current_position,
                    piece->destinations[destination_index].destination);
                if (new_distance >
                    piece->destinations[destination_index].distance)
                {
                    distance_t delta = new_distance -
                        piece->destinations[destination_index].distance;
                    if ( ( piece->camp == BLANC &&
                           delta > NbCoupsBlancsRestants ) ||
                         ( piece->camp == NOIR  &&
                           delta > NbCoupsNoirsRestants ) )
                    {
/*  fprintf(MainFD, */
/*  	"Position OUT a cause de la piece : 0x%02x. Type = %d(%d).\n", */
/*  	piece->caseInitiale, piece->typePiece, type_piece); */
/*  fprintf(MainFD, */
/*  	"current_position = 0x%02x. destination = 0x%02x. ", */
/*  	current_position, */
/*  	piece->destination[destination_index] ); */
/*  fprintf(MainFD, "distance = %d. Nouvelle distance = %d\n", */
/*  	piece->distance[destination_index], */
/*  	new_distance); */
/*  writeAssociation( "" ); */
                        distance_pop();
                        for (i=0; i<nb_new_distances; i++)
                        {
                            ListeSures[new_distances[i].piece_indice]
                                .destinations[new_distances[i].destination_index]
                                .distance =
                                new_distances[i].previous_distance;
                        }
                        NbCoupsBlancsRestants = white_spare_moves;
                        NbCoupsNoirsRestants  = black_spare_moves;
                        for (i=0; i<nb_new_forbidden_squares; i++)
                        {
                            CasesInterdites[new_forbidden_squares[i]] = FALSE;
                        }
                        return;
                    }
                    if (nb_new_distances < max_nb_of_new_distances)
                    {
                        new_distances[nb_new_distances].piece_indice = indice;
                        new_distances[nb_new_distances].destination_index =
                            destination_index;
                        new_distances[nb_new_distances].previous_distance =
                            piece->destinations[destination_index].distance;
                        piece->destinations[destination_index].distance =
                            new_distance;
                        nb_new_distances++;

                        if (piece->camp == BLANC)
                        {
                            NbCoupsBlancsRestants -= delta;
                        }
                        else
                        {
                            NbCoupsNoirsRestants  -= delta;
                        }
                    }
                    else
                    {
                        fprintf(MainFD,
                                "Warning : maximum number of new distances reached in update_trajectories_due_to_stationary_pieces\n");
                    }
                }
            }
            current_position =
                piece->destinations[destination_index].destination;
        }
    }

    precise_trajectories( 0, 0 );

    distance_pop();

    for (i=0; i<nb_new_distances; i++)
    {
        ListeSures[new_distances[i].piece_indice]
            .destinations[new_distances[i].destination_index].distance =
            new_distances[i].previous_distance;
    }
    NbCoupsBlancsRestants = white_spare_moves;
    NbCoupsNoirsRestants  = black_spare_moves;

    for (i=0; i<nb_new_forbidden_squares; i++)
    {
        CasesInterdites[new_forbidden_squares[i]] = FALSE;
    }
}

static bool_t
is_pawn_trajectory_conflicting_with_twin_square(
    const twin_squares_t* twin_squares,
    square_t pawn_origin,
    square_t current,
    square_t dest,
    piece_index_t capturedPiece,
    int deltaY)
{
    if (capturedPiece == PAS_DE_CAPTURE)
    {
        while (current != dest)
        {
            square_t twin = find_twin_square(
                twin_squares, current, e_vertical, FOU, pawn_origin);

            if ((deltaY > 0 && twin > current && twin <= dest)
                || (twin < current && twin >= dest))
                return TRUE;

            current += deltaY;
        }
    }
    else
    {
        square_t twin = find_twin_square(
            twin_squares, current, e_vertical, TOUR, pawn_origin);
        if (twin == dest) return TRUE;
    }
    return FALSE;
}
